#ifdef CURSES
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <ncurses.h>
#include "laroec.h"
#include "xmalloc.h"


int log_len = 0;
static char **session_log;
 
static void
background_change(WINDOW *w, char *s, int attr) 
{
        wattrset(w, attr);
        waddstr(w, s);
        wrefresh(w);
}

static inline void 
select_menuitem(WINDOW *w, char *s)
{
        background_change(w, s, A_REVERSE);
}

static inline void 
unselect_menuitem(WINDOW *w, char *s)
{
        background_change(w, s, A_NORMAL);
}

static void
clear_menus(WINDOW **menuitems, int x)
{
        int i;
        for(i=0; i < x; i++) {
                werase(menuitems[i]);
                wrefresh(menuitems[i]);
        }
}

static void
draw_menus(WINDOW **menuitems, char **menutext, int x)
{
        int i;
        for(i=0; i<x; i++) {
                waddstr(menuitems[i], menutext[i]);
                wrefresh(menuitems[i]);
        }
}

static void
delete_menus(WINDOW **menuitems, int x)
{
     int i;
     clear_menus(menuitems, x);
     for(i=0; i<x; i++) 
	  delwin(menuitems[i]);
}
	  


static WINDOW *l_menus[3];
static char *l_mtext[] = {
        "Menu",
        "Log", 
        "Scroll",
        "Quit"
};


static void
setup_laroec_menus()
{
        int offset, i;
        offset = ((COLS * .8) / 4);
        for(i=0; i < 4; i++) {
                l_menus[i] = newwin(1, strlen(l_mtext[i]), LINES - 1, offset * i);
                waddstr(l_menus[i], l_mtext[i]);
                keypad(l_menus[i], TRUE);
                wnoutrefresh(l_menus[i]);
        }
        doupdate();
}

void
draw_screen_backwards(int fl, int ll, WINDOW *win, char **text, int lt)
{
     werase(win);

     while(ll >= fl) {
	  mvwaddstr(win, ll, 0, text[lt]);
	  ll--; lt--;
     }
     wrefresh(win);
}

static void
do_scroll(WINDOW *win)
{
     int c;
     int maxx, maxy;
     WINDOW *scrollmsg;
     char last_input[1024];
     int log_pos = log_len;

     getmaxyx(win, maxy, maxx);

     clear_menus(l_menus, 4);

     scrollmsg = newwin(1, COLS - 1, LINES - 1, 0);

     if ( log_len <= maxy ) {
	  waddstr(scrollmsg, "There's nothing to scroll, bucko.");
	  wrefresh(scrollmsg);
	  usleep(500000);
	  
	  wclear(scrollmsg); wrefresh(scrollmsg); delwin(scrollmsg);

	  draw_menus(l_menus, l_mtext, 4); curs_set(TRUE); noecho(); 
	  return;
     }
 
     waddstr(scrollmsg, "Use UP and DOWN ARROR to scroll and ESCAPE to return to laroec.");
     wrefresh(scrollmsg);
     keypad(scrollmsg, TRUE);

     mvwinstr(win, maxy - 1, 0, last_input);

     while((c = wgetch(scrollmsg))) {
	  switch(c) {
	  case KEY_UP:
	       if ( log_pos <= maxy )
		    break;
	       draw_screen_backwards(0, maxy, win, session_log, --log_pos);
	       break;
	  case KEY_DOWN:
	       if ( log_pos == log_len )
		    break;
	       draw_screen_backwards(0, maxy, win, session_log, ++log_pos);
	       break;
	  case 27:
	       draw_screen_backwards(0, maxy - 1, win, session_log, log_len);
	       mvwaddstr(win, maxy - 1, 0, last_input);

	       wclear(scrollmsg);
	       wrefresh(scrollmsg);
	       delwin(scrollmsg);

	       draw_menus(l_menus, l_mtext, 4);
	       curs_set(TRUE);
	       noecho();
	       return;
	  }
     }
}


static void
do_log()
{
        char **p;
        char fn[1024];
        FILE *fp;
        WINDOW *lgwin;

        echo();
        clear_menus(l_menus, 4);
        curs_set(TRUE);

        lgwin = newwin(1, COLS - 1, LINES - 1, 0);
        waddstr(lgwin, "File to log to: ");
        wrefresh(lgwin);

        wgetnstr(lgwin, fn, 1024);

        curs_set(FALSE);

        fp = fopen(fn, "a");
        for(p = session_log; *p != (char *) NULL; p++)
                fprintf(fp, "%s\n", *p);
        fclose(fp);

        wprintw(lgwin, "Session logged to %s", fn);
        wrefresh(lgwin);

        usleep(500000);

        wclear(lgwin);
        wrefresh(lgwin);
        delwin(lgwin);

        draw_menus(l_menus, l_mtext, 4);
        curs_set(TRUE);
        noecho();
}

static int 
do_menus(WINDOW *mw)
{
        int c;
        static int current;
        curs_set(FALSE);
        select_menuitem(l_menus[current], l_mtext[current]);
        while ((c = wgetch(l_menus[current]))) {
                switch(c) {
                case '\r':
		case '\n':
                        switch(current) {
			case 0:
			     delete_menus(l_menus, 4);
			     return(0);
                        case 1:
			     unselect_menuitem(l_menus[current], l_mtext[current]);
			     do_log();
			     return(1);
			case 2:
			     unselect_menuitem(l_menus[current], l_mtext[current]);
			     do_scroll(mw);
			     return(1);
                        case 3:
			     clear_menus(l_menus, 4);
			     endwin();
			     exit(0);
                        }
                        break;
                case KEY_UP:
                        unselect_menuitem(l_menus[current], l_mtext[current]);
                        curs_set(TRUE);
                        return(1);
                case KEY_LEFT:
                        unselect_menuitem(l_menus[current], l_mtext[current]);
                        current = current ? current - 1 : 3;
                        select_menuitem(l_menus[current], l_mtext[current]);
                        break;
                case KEY_RIGHT:
                        unselect_menuitem(l_menus[current], l_mtext[current]);
                        current = (current < 3) ? current + 1 : 0;
                        select_menuitem(l_menus[current], l_mtext[current]);
                        break;
                }
        }
	return(1);
}
                                

static int count_words(char *str) 
{
        int i;
        while( isspace(*str) ) str++;
        for(i = 0; *str; str++) 
                if ( isspace(*str) ) i++;
        return(i);
}

static void rand_delay(char c) 
{
        if ( isspace(c) ) 
                usleep((long) (((double) 300000) * (rand()/(double) RAND_MAX)));
        else 
                usleep(50000);
}        

static void typewrite_sentence(char *str, WINDOW *win)
{
        int bold, word_bold, i;
        if ( (rand()/(double) RAND_MAX) < .2 ) {
                word_bold = (int) ((double) count_words(str)) * (rand()/(double) RAND_MAX);
                for(i=0, bold = 0; *str; str++) {
                        if ( isspace(*str) ) i++;
                        if ( i == word_bold ) {
                                bold = 1;
                                wattrset(win, A_BOLD);
                        } else if ( i > word_bold && bold ) {
                                bold = 0;
                                wattrset(win, A_NORMAL);
                        } 
                        rand_delay(*str);
                        waddch(win, *str);
                        wrefresh(win);
                }
        } else {
                while ( *str ) {
                        rand_delay(*str);
                        waddch(win, *str++);
                        wrefresh(win);
                }
        }
}
 
static char *
do_input(WINDOW *win)
{
        int c, x, y, i, j;
        static char buf[1024];
        noecho();
        nonl();
        i = j = 0;
        while ((c = wgetch(win))){
                if ( c == '\r' ) {
                        waddch(win, '\n');
                        wrefresh(win);
                        buf[j] = '\0';
                        return(buf);
                } else if ( c == KEY_DOWN ) {
                        getyx(win, y, x);
                        if ( !do_menus(win) )
			     return(NULL);
                        wmove(win, y, x);
                } else if ( c == erasechar() ) {
                        if ( i ) {
                                getyx(win, y, x);
                                mvwdelch(win, y, x-1);
                                wrefresh(win);
                                i--; j--;
                        }
                } else {
                        buf[j++] = c;
                        i++; 
                        waddch(win, c);
                        wrefresh(win);
                }
        }
	return(NULL);
}
                
static void set_user_question_attributes(WINDOW *win)
{
        curs_set(TRUE);
        echo();
        wattrset(win, A_NORMAL);
}


static void set_laroce_answer_attributes(WINDOW *win)
{
        curs_set(FALSE);
        noecho();
}

static void cp_to_log(char *str, int n)
{
     session_log = realloc(session_log, sizeof(char *) * (log_len + 2));
     if ( n < 0 ) {
	  session_log[log_len] = xmalloc(strlen(str) + 1);
	  strcpy(session_log[log_len], str);
     } else {
	  session_log[log_len] = xmalloc(n + 1);
	  strncpy(session_log[log_len], str, n);
	  session_log[log_len][n] = '\0';
     }
     session_log[++log_len] = (char *) NULL;
}

static void add_to_log(char *fmt, ...) 
{
        static char buf[1024];
        char *p = buf;
	char *s = buf;

        va_list ap;
        va_start(ap, fmt);
        vsprintf(buf, fmt, ap);
        
	if ( index(p, '\n') == NULL ) {
	     cp_to_log(p, -1);
	} else {
	     while ( (p = index(s, '\n')) != NULL ) {
		  cp_to_log(s, p - s);
		  s = ++p;
	     }
	     cp_to_log(s, -1);
	}
}

static void clear_log()
{
     int i;
     for(i=0; i < log_len; i++)
	  free(session_log[i]);
     free(session_log);
     log_len = 0;
     session_log = NULL;
}

static void
do_ifthen() 
{
        char *p, *i;
        WINDOW *win, *border;

        setup_laroec_menus();

        /* set up the border */
        border = newwin(LINES -1, COLS, 0, 0);
        box(border, 0, 0);        
        wrefresh(border);
        
        win = newwin(LINES - 5, COLS - 6, 2, 3);
        scrollok(win, TRUE);
        keypad(win, TRUE);

	initialize_laroec(TYPE_IFTHEN);
        
        add_to_log("--");
        wprintw(win, "--\n");
        set_user_question_attributes(win);
        while ( 1 ) {
                wprintw(win, "If ");
                wrefresh(win);

		i = do_input(win);
		if ( i == (char *) NULL ) {
		     delwin(win); 
		     delwin(border);
		     clear_log();
		     return;
		} 
		
		add_to_log("If %s", i);

                set_laroce_answer_attributes(win);

                wprintw(win, "then ");
                p = get_random_phrase();
                typewrite_sentence(p, win);

                add_to_log("then %s", p);

                set_user_question_attributes(win);
                wprintw(win, "\n--\n");
                add_to_log("--");
        } 
}

static void
do_comesfrom()
{
     char *p, *i;
     WINDOW *win, *border;

     setup_laroec_menus();
     
     /* set up the border */
     border = newwin(LINES -1, COLS, 0, 0);
     box(border, 0, 0);        
     wrefresh(border);
     
     win = newwin(LINES - 5, COLS - 6, 2, 3);
     scrollok(win, TRUE);
     keypad(win, TRUE);
     
     initialize_laroec(TYPE_COMESFROM);

     add_to_log("--");
     wprintw(win, "--\n");
     set_user_question_attributes(win);
     while ( 1 ) {
	  i = do_input(win);
	  
	  if ( i == (char *) NULL ) {
	       delwin(win); 
	       delwin(border);
	       clear_log();
	       return;
	  }

	  add_to_log("%s", i);
	  
	  set_laroce_answer_attributes(win);
	  
	  wprintw(win, "Comes from ");
	  p = get_random_phrase();
	  typewrite_sentence(p, win);
	  
	  add_to_log("Comes from %s", p);

	  set_user_question_attributes(win);
	  wprintw(win, "\n--\n");
	  add_to_log("--");
     } 
}

/*
 * Begin curses intro functions 
 */


static void
post_char(char c, WINDOW *win, int line, int col, int offset)
{
     mvwaddch(win, line, col + offset, c);
     wrefresh(win);
}

#define DC '0'
#define TOP_TITLE "Links At Random"
#define BOTTOM_TITLE "Order Emerges from Chaos"

static int
choose_random_char(char *s)
{
     char i = DC;
     int c;
     int len = strlen(s);
     while(i == DC) {
	  c = (int) (len * (rand() / (double) RAND_MAX));
	  i = s[c];
     }
     return(c);
}

static void
spiff_menu_title(WINDOW *win, int ts, int bs, int bl)
{
     int pos;

     static int ntop_done = 0;
     static int nbottom_done = 0;

     static int top_done = 0;
     static int bottom_done = 0;
     static int done = 0;
     static char top[] = TOP_TITLE;
     static char bottom[] = BOTTOM_TITLE;

     int top_length = strlen(top);
     int bottom_length = strlen(bottom);

     /* what a hack this is turning out to be.  If win == NULL and everything else 
	is -1, reset ourselves. */
     if ( win == NULL && ts < 0 && bs < 0 && bl < 0 ) {
	  done = ntop_done = nbottom_done = top_done = bottom_done = 0;
	  strcpy(top, TOP_TITLE);
	  strcpy(bottom, BOTTOM_TITLE);
	  return;
     }
     
     if ( done )
	  return;

     if ( top_done ) {
	  pos = choose_random_char(bottom);
	  post_char(bottom[pos], win, bl, bs, pos);
	  bottom[pos] = DC;
	  nbottom_done++;
     } else if ( bottom_done ) {
	  pos = choose_random_char(top);
	  post_char(top[pos], win, 0, ts, pos);
	  ntop_done++;
	  top[pos] = DC;
     } else {
	  if ( (rand() / (double) RAND_MAX) < .5 ) {
	       pos = choose_random_char(bottom);
	       post_char(bottom[pos], win, bl, bs, pos);
	       bottom[pos] = DC;
	       nbottom_done++;
	  } else {
	       pos = choose_random_char(top);
	       post_char(top[pos], win, 0, ts, pos);
	       top[pos] = DC;
	       ntop_done++;
	  }
     }
     if ( ntop_done == top_length )
	  top_done = 1;
     if ( nbottom_done == bottom_length )
	  bottom_done = 1;
     if ( top_done && bottom_done )
	  done = 1;
}

static char *opening_menu_text[] = {
     "If X then Y",
     "X Comes from Y",
     "Quit"
};

static WINDOW **
setup_opening_menus(WINDOW *win, int y, int x)
{
     static WINDOW *menus[2];

     int i, j;
     for(i = ((y - 2) / 2), j=0; j < 3; i++, j++) {
	  menus[j] = derwin(win, 1, strlen(opening_menu_text[j]), i,
			    (x - strlen(opening_menu_text[j])) / 2);
	  waddstr(menus[j], opening_menu_text[j]);
	  wrefresh(menus[j]);
	  nodelay(menus[j], TRUE);
	  keypad(menus[j], TRUE);
     }
     return(menus);
} 

void
reset_opening_menus(WINDOW *w, WINDOW **m, int x)
{
     erase();
     refresh();
     
     box(w, 0, 0);
     wrefresh(w);

     draw_menus(m, opening_menu_text, 3);
     curs_set(FALSE);
     
     select_menuitem(m[x], opening_menu_text[x]);

     spiff_menu_title(NULL, -1, -1, -1);
}

static int
opening_menu()
{
     int mw_h, mw_w;
     int cur = 0, f = 1;
     WINDOW *mw, **menus;
     mw_w = COLS * .5;
     mw_h = LINES * .3;

     mw = newwin(mw_h, mw_w, (LINES - mw_h) / 2, (COLS - mw_w) / 2);
     box(mw, 0, 0);
     wrefresh(mw);

     srand(time(NULL));
     menus = setup_opening_menus(mw, mw_h, mw_w); 
     select_menuitem(menus[0], opening_menu_text[0]);
     curs_set(FALSE);

     while(1) {
	  if ( f ) 
	       spiff_menu_title(mw, 5, 5, mw_h - 1);
	  switch(wgetch(menus[cur])) {
	  case ERR:
	       break;
	  case KEY_UP:
	       unselect_menuitem(menus[cur], opening_menu_text[cur]);
	       cur = cur ? cur - 1 : 2;
	       select_menuitem(menus[cur], opening_menu_text[cur]);
	       break;
	  case KEY_DOWN:
	       unselect_menuitem(menus[cur], opening_menu_text[cur]);
	       cur = cur == 2 ? 0 : cur + 1;
	       select_menuitem(menus[cur], opening_menu_text[cur]);
	       break;
	  case '\r':
	  case '\n':
	       switch(cur) {
	       case 0:
		    do_ifthen();
		    reset_opening_menus(mw, menus, cur);
		    break;
	       case 1:
		    do_comesfrom();
		    reset_opening_menus(mw, menus, cur);
		    break;
	       case 2:
		    endwin();
		    exit(0);
	       }
	  }
	  usleep(80000);
	  f = f ? 0 : 1;
     }
     
     for(f = 0; f < 3; f++) 
	  delwin(menus[f]);
     delwin(mw); 
     return(1);
}

/* 
 * End curses intro functions
 */



int
curses_ui_main(int argc, char **argv)
{
        initscr();

        start_color();
        cbreak();
        noecho();

	opening_menu(); 
        endwin();
	return(1);
}

#endif













